  /*
     Driver for Philips tda1004xh OFDM Frontend

     (c) 2003, 2004 Andrew de Quincey & Robert Schlabbach

     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   */
/*
 * This driver needs external firmware. Please use the commands
 * "<kerneldir>/Documentation/dvb/get_dvb_firmware tda10045",
 * "<kerneldir>/Documentation/dvb/get_dvb_firmware tda10046" to
 * download/extract them, and then copy them to /usr/lib/hotplug/firmware.
 */
#define TDA10045_DEFAULT_FIRMWARE "dvb-fe-tda10045.fw"
#define TDA10046_DEFAULT_FIRMWARE "dvb-fe-tda10046.fw"

#include <linux/init.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/device.h>
#include "dvb_frontend.h"
#include "tda1004x.h"



#define FRONTEND_NAME "dvbfe_tda1004x"

#define dprintk(args...) \
	do { \
		if (debug) printk(KERN_DEBUG FRONTEND_NAME ": " args); \
	} while (0)


static int debug;
module_param(debug, int, 0644);
MODULE_PARM_DESC(debug, "Turn on/off frontend debugging (default:off).");


#define TDA1004X_CHIPID		 0x00
#define TDA1004X_AUTO		 0x01
#define TDA1004X_IN_CONF1	 0x02
#define TDA1004X_IN_CONF2	 0x03
#define TDA1004X_OUT_CONF1	 0x04
#define TDA1004X_OUT_CONF2	 0x05
#define TDA1004X_STATUS_CD	 0x06
#define TDA1004X_CONFC4		 0x07
#define TDA1004X_DSSPARE2	 0x0C
#define TDA10045H_CODE_IN	 0x0D
#define TDA10045H_FWPAGE	 0x0E
#define TDA1004X_SCAN_CPT	 0x10
#define TDA1004X_DSP_CMD	 0x11
#define TDA1004X_DSP_ARG	 0x12
#define TDA1004X_DSP_DATA1	 0x13
#define TDA1004X_DSP_DATA2	 0x14
#define TDA1004X_CONFADC1	 0x15
#define TDA1004X_CONFC1		 0x16
#define TDA10045H_S_AGC		 0x1a
#define TDA10046H_AGC_TUN_LEVEL	 0x1a
#define TDA1004X_SNR		 0x1c
#define TDA1004X_CONF_TS1	 0x1e
#define TDA1004X_CONF_TS2	 0x1f
#define TDA1004X_CBER_RESET	 0x20
#define TDA1004X_CBER_MSB	 0x21
#define TDA1004X_CBER_LSB	 0x22
#define TDA1004X_CVBER_LUT	 0x23
#define TDA1004X_VBER_MSB	 0x24
#define TDA1004X_VBER_MID	 0x25
#define TDA1004X_VBER_LSB	 0x26
#define TDA1004X_UNCOR		 0x27

#define TDA10045H_CONFPLL_P	 0x2D
#define TDA10045H_CONFPLL_M_MSB	 0x2E
#define TDA10045H_CONFPLL_M_LSB	 0x2F
#define TDA10045H_CONFPLL_N	 0x30

#define TDA10046H_CONFPLL1	 0x2D
#define TDA10046H_CONFPLL2	 0x2F
#define TDA10046H_CONFPLL3	 0x30
#define TDA10046H_TIME_WREF1	 0x31
#define TDA10046H_TIME_WREF2	 0x32
#define TDA10046H_TIME_WREF3	 0x33
#define TDA10046H_TIME_WREF4	 0x34
#define TDA10046H_TIME_WREF5	 0x35

#define TDA10045H_UNSURW_MSB	 0x31
#define TDA10045H_UNSURW_LSB	 0x32
#define TDA10045H_WREF_MSB	 0x33
#define TDA10045H_WREF_MID	 0x34
#define TDA10045H_WREF_LSB	 0x35
#define TDA10045H_MUXOUT	 0x36
#define TDA1004X_CONFADC2	 0x37

#define TDA10045H_IOFFSET	 0x38

#define TDA10046H_CONF_TRISTATE1 0x3B
#define TDA10046H_CONF_TRISTATE2 0x3C
#define TDA10046H_CONF_POLARITY	 0x3D
#define TDA10046H_FREQ_OFFSET	 0x3E
#define TDA10046H_GPIO_OUT_SEL	 0x41
#define TDA10046H_GPIO_SELECT	 0x42
#define TDA10046H_AGC_CONF	 0x43
#define TDA10046H_AGC_GAINS	 0x46
#define TDA10046H_AGC_TUN_MIN	 0x47
#define TDA10046H_AGC_TUN_MAX	 0x48
#define TDA10046H_AGC_IF_MIN	 0x49
#define TDA10046H_AGC_IF_MAX	 0x4A

#define TDA10046H_FREQ_PHY2_MSB	 0x4D
#define TDA10046H_FREQ_PHY2_LSB	 0x4E

#define TDA10046H_CVBER_CTRL	 0x4F
#define TDA10046H_AGC_IF_LEVEL	 0x52
#define TDA10046H_CODE_CPT	 0x57
#define TDA10046H_CODE_IN	 0x58


#define FE_TYPE_TDA10045H     0
#define FE_TYPE_TDA10046H     1



static u8 tda10045_inittab[] = {
	TDA1004X_CONFC4, 0,
	TDA1004X_AUTO, 0x1f,
	TDA1004X_CONFC1, 0x88,
	TDA1004X_IN_CONF2, 0x00,
	TDA1004X_CONF_TS1, 0,
	TDA1004X_CONF_TS2, 0,
	TDA1004X_VBER_MSB, 0xa0,
	TDA1004X_CONFADC1, 0x2e,
};

static u8 tda10046_inittab[] = { 
	TDA1004X_CONFC4, 0x20, 0,          // disable DSP watchdog timer
	TDA1004X_CONFC1, 0x08,
	TDA1004X_AUTO, 8, 0,               // select HP stream
	TDA10046H_CONFPLL2, 10,            // PLL M = 10   
	TDA10046H_CONFPLL3, 0,             // PLL P = N = 0	  
	TDA10046H_FREQ_OFFSET, 99,         // FREQOFFS = 99
	TDA10046H_FREQ_PHY2_MSB, 0xd4,     // } PHY2 = -11221
	TDA10046H_FREQ_PHY2_LSB, 0x2c,     // }
	TDA10046H_AGC_CONF, 0,             // AGC setup
	TDA10046H_CONF_POLARITY, 0x60,     // set AGC polarities
	TDA10046H_AGC_TUN_MIN, 0,          // }	  
	TDA10046H_AGC_TUN_MAX, 0xff,       // } AGC min/max values
	TDA10046H_AGC_IF_MIN, 0,           // }
	TDA10046H_AGC_IF_MAX, 0xff,        // }
	TDA10046H_CVBER_CTRL, 0x1a,        // 10^6 VBER measurement bits	  
	TDA10046H_AGC_GAINS, 1,            // IF gain 2, TUN gain 1
	TDA1004X_AUTO, 0x00,               // crystal is 50ppm	
	TDA1004X_CONF_TS1, 7,             // MPEG2 interface config
	TDA1004X_CONF_TS2, 0x31, 0,        // MPEG2 interface config
	TDA10046H_CONF_TRISTATE1, 0x00,    // disable AGC_TUN	
	TDA10046H_CONF_TRISTATE2, 0xe1,    // tristate setup
	TDA10046H_GPIO_OUT_SEL, 0xcc,      // GPIO output config
	TDA10046H_GPIO_SELECT, 8,          // GPIO select
};

   

static struct dvb_frontend_info tda10045h_info = {
	.name = "Philips TDA10045H",
	.type = FE_OFDM,
	.frequency_min = 51000000,
	.frequency_max = 858000000,
	.frequency_stepsize = 166667,
	.caps =
	    FE_CAN_FEC_1_2 | FE_CAN_FEC_2_3 | FE_CAN_FEC_3_4 |
	    FE_CAN_FEC_5_6 | FE_CAN_FEC_7_8 | FE_CAN_FEC_AUTO |
	    FE_CAN_QPSK | FE_CAN_QAM_16 | FE_CAN_QAM_64 | FE_CAN_QAM_AUTO |
	    FE_CAN_TRANSMISSION_MODE_AUTO | FE_CAN_GUARD_INTERVAL_AUTO
};

static struct dvb_frontend_info tda10046h_info = {
	.name = "Philips TDA10046H",
	.type = FE_OFDM,
	.frequency_min = 51000000,
	.frequency_max = 858000000,
	.frequency_stepsize = 166667,
	.caps =
	    FE_CAN_FEC_1_2 | FE_CAN_FEC_2_3 | FE_CAN_FEC_3_4 |
	    FE_CAN_FEC_5_6 | FE_CAN_FEC_7_8 | FE_CAN_FEC_AUTO |
	    FE_CAN_QPSK | FE_CAN_QAM_16 | FE_CAN_QAM_64 | FE_CAN_QAM_AUTO |
	    FE_CAN_TRANSMISSION_MODE_AUTO | FE_CAN_GUARD_INTERVAL_AUTO
};

static int tda1004x_write_mask(struct tda1004x_cardspec* cardspec, int reg, int mask, int data)
{
	u8 val;
	int result;

	dprintk("%s: reg=0x%x, mask=0x%x, data=0x%x\n", __FUNCTION__, reg,
		mask, data);

	// read a byte and check
	if ((result = cardspec->demod_read(cardspec, reg, &val, 1)) !=0) return result;

	// mask if off
	val = val & ~mask;
	val |= data & 0xff;

	// write it out again
	return cardspec->demod_write(cardspec, reg, &val, 1);
}

static int tda1004x_write_buf(struct tda1004x_cardspec* cardspec, int reg, unsigned char *buf, int len)
{
	int i;
	int result;

	dprintk("%s: reg=0x%x, len=0x%x\n", __FUNCTION__, reg, len);

	result = 0;
	for (i = 0; i < len; i++) {
		if ((result = cardspec->demod_write(cardspec, reg+i, buf+i, 1)) !=0) return result;
	}

	return 0;
}

static int tda1004x_enable_tuner_i2c(struct tda1004x_cardspec* cardspec)
{
	int result;
	dprintk("%s\n", __FUNCTION__);

	if ((result = tda1004x_write_mask(cardspec, TDA1004X_CONFC4, 2, 2)) !=0) return result;
	msleep(1);
	return 0;
}

static int tda1004x_disable_tuner_i2c(struct tda1004x_cardspec* cardspec)
{
	dprintk("%s\n", __FUNCTION__);

	return tda1004x_write_mask(cardspec, TDA1004X_CONFC4, 2, 0);
}

static int tda10045h_set_bandwidth(struct tda1004x_cardspec* cardspec,
				   fe_bandwidth_t bandwidth)
{
	static u8 bandwidth_6mhz[] = { 0x02, 0x00, 0x3d, 0x00, 0x60, 0x1e, 0xa7, 0x45, 0x4f };
	static u8 bandwidth_7mhz[] = { 0x02, 0x00, 0x37, 0x00, 0x4a, 0x2f, 0x6d, 0x76, 0xdb };
	static u8 bandwidth_8mhz[] = { 0x02, 0x00, 0x3d, 0x00, 0x48, 0x17, 0x89, 0xc7, 0x14 };
	int result;
	u8 tmp = 0;
	u8 dsspare;
	u8* bw_settings;

	switch (bandwidth) {
	case BANDWIDTH_6_MHZ:
		dsspare = 0x14;
	   	bw_settings = bandwidth_6mhz;
		break;

	case BANDWIDTH_7_MHZ:
		dsspare = 0x80;
		bw_settings = bandwidth_7mhz;
		break;

	case BANDWIDTH_8_MHZ:
		dsspare = 0x14;
   		bw_settings = bandwidth_8mhz;
		break;

	default:
		return -EINVAL;
	}

	if ((result = cardspec->demod_write(cardspec, TDA1004X_DSSPARE2, &dsspare, 1)) !=0) return result;
	if ((result = tda1004x_write_buf(cardspec, TDA10045H_CONFPLL_P, bw_settings, sizeof(bandwidth_6mhz))) !=0) return result;
	if ((result = cardspec->demod_write(cardspec, TDA10045H_IOFFSET, &tmp, 1)) !=0) return result;

	return 0;
}

static int tda10046h_set_bandwidth(struct tda1004x_cardspec* cardspec,
				   fe_bandwidth_t bandwidth)
{
	static u8 bandwidth_6mhz[] = { 0x80, 0x15, 0xfe, 0xab, 0x8e };
	static u8 bandwidth_7mhz[] = { 0x6e, 0x02, 0x53, 0xc8, 0x25 };
	static u8 bandwidth_8mhz[] = { 0x60, 0x12, 0xa8, 0xe4, 0xbd };
	int result;
	u8 dsspare;
	u8* bw_settings;

	switch (bandwidth) {
	case BANDWIDTH_6_MHZ:
		dsspare = 0;
		bw_settings = bandwidth_6mhz;
		break;

	case BANDWIDTH_7_MHZ:
		dsspare = 0;
		bw_settings = bandwidth_7mhz;
		break;

	case BANDWIDTH_8_MHZ:
		dsspare = 0xff;
		bw_settings = bandwidth_8mhz;
		break;

	default:
		return -EINVAL;
	}

	if ((result = tda1004x_write_buf(cardspec, TDA10046H_TIME_WREF1, bw_settings, sizeof(bandwidth_6mhz))) !=0) return result;
	if ((result = cardspec->demod_write(cardspec, TDA1004X_DSSPARE2, &dsspare, 1)) !=0) return result;

	return 0;
}

static int tda1004x_do_upload(struct tda1004x_cardspec* cardspec, unsigned char *mem, unsigned int len,
			      u8 code_counter_reg, u8 code_in_reg)
{
	u8 buf[0x10];
	int result;
	int tx_size;
	int pos = 0;
	u8 tmp = 0;

	/* clear code counter */
	if ((result = cardspec->demod_write(cardspec, code_counter_reg, &tmp, 1)) !=0) return result;

	while (pos != len) {

		// work out how much to send this time
		tx_size = len - pos;
		if (tx_size > 0x10) {
			tx_size = 0x10;
		}

		// send the chunk
		memcpy(buf, mem + pos, tx_size);
		if ((result = cardspec->demod_write(cardspec, code_in_reg, buf, tx_size)) !=0) return result;
		pos += tx_size;

		dprintk("%s: fw_pos=0x%x\n", __FUNCTION__, pos);
	}
	return 0;
}

static int tda1004x_check_upload_ok(struct tda1004x_cardspec* cardspec, u8 dsp_version)
{
	int result;
	u8 cmd = 0x67;
	u8 data1, data2;

	// issue command
	if ((result = tda1004x_write_mask(cardspec, TDA1004X_CONFC4, 0x10, 0)) !=0) return result;
	if ((result = cardspec->demod_write(cardspec, TDA1004X_DSP_CMD, &cmd, 1)) !=0) return result;

	// check results
	if ((result = cardspec->demod_read(cardspec, TDA1004X_DSP_DATA1, &data1, 1)) !=0) return result;
	if ((result = cardspec->demod_read(cardspec, TDA1004X_DSP_DATA2, &data2, 1)) !=0) return result;
	if ((data1 != 0x67) || (data2 != dsp_version)) return -EIO;

	return 0;
}


static int tda10045_fwupload(struct tda1004x_cardspec* cardspec)
{
	int result;
	const struct firmware *fw;

	/* request the firmware, this will block until someone uploads it */
	printk("tda1004x: waiting for firmware upload...\n");
	if ((result = cardspec->request_firmware(cardspec, &fw, TDA10045_DEFAULT_FIRMWARE)) !=0) {
		printk("tda1004x: no firmware upload (timeout or file not found?)\n");
	   	return result;
	}
   
	/* set some valid bandwith parameters before uploading */

	/* reset chip */
	if ((result = tda1004x_write_mask(cardspec, TDA1004X_CONFC4, 0x10, 0)) !=0) return result;
	if ((result = tda1004x_write_mask(cardspec, TDA1004X_CONFC4, 8, 8)) !=0) return result;
	if ((result = tda1004x_write_mask(cardspec, TDA1004X_CONFC4, 8, 0)) !=0) return result;
	msleep(10);

	/* set parameters */
	if ((result = tda10045h_set_bandwidth(cardspec, BANDWIDTH_8_MHZ)) !=0) return result;
	if ((result = tda1004x_do_upload(cardspec, fw->data, fw->size, TDA10045H_FWPAGE, TDA10045H_CODE_IN)) !=0) return result;

	/* wait for DSP to initialise */
	/* DSPREADY doesn't seem to work on the TDA10045H */
	msleep(100);

	if ((result = tda1004x_check_upload_ok(cardspec, 0x2c)) !=0) return result;

	return 0;
}

static int tda10046_fwupload(struct tda1004x_cardspec* cardspec)
{
	unsigned long timeout;
	int result;
	const struct firmware *fw;
   	u8 val;

	/* request the firmware, this will block until someone uploads it */
	printk("tda1004x: waiting for firmware upload...\n");
	if ((result = cardspec->request_firmware(cardspec, &fw, TDA10046_DEFAULT_FIRMWARE)) !=0) {
		printk("tda1004x: no firmware upload (timeout or file not found?)\n");
   	   	return result;
	}

	/* set some valid bandwith parameters before uploading */

	/* reset chip */
	if ((result = tda1004x_write_mask(cardspec, TDA1004X_CONFC4, 1, 0)) !=0) return result;
	if ((result = tda1004x_write_mask(cardspec, TDA10046H_CONF_TRISTATE1, 1, 0)) !=0) return result;
	if ((result = tda1004x_write_mask(cardspec, TDA1004X_CONFC4, 8, 8)) !=0) return result;
	msleep(10);

	if ((result = tda1004x_do_upload(cardspec, fw->data, fw->size, TDA10046H_CODE_CPT, TDA10046H_CODE_IN)) !=0) return result;

	/* wait for DSP to initialise */
	timeout = jiffies + HZ;
	while(1) {
	   	/* has it finished yet? */
		if ((result = cardspec->demod_read(cardspec, TDA1004X_STATUS_CD, &val, 1)) !=0) return result;
		if (val & 0x20) break;

		/* try try again */
		if (time_after(jiffies, timeout)) {
			printk("tda1004x: DSP failed to initialised.\n");
			return -EIO;
		}
		msleep(1);
	}

	if ((result = tda1004x_check_upload_ok(cardspec, 0x20)) !=0) return result;

	return 0;
}

static int tda1004x_init(struct tda1004x_cardspec* cardspec)
{
	u8* inittab;
	int initcount;
	int i;
	int result;
   
	dprintk("%s\n", __FUNCTION__);

	switch(cardspec->demod_type) {
	case FE_TYPE_TDA10045H:
		if ((result = tda1004x_write_mask(cardspec, TDA1004X_CONFADC1, 0x10, 0)) !=0) return result; // wake up the ADC
	   	inittab = tda10045_inittab;
		initcount = sizeof(tda10045_inittab) / 2;
		break;

	case FE_TYPE_TDA10046H:
		if ((result = tda1004x_write_mask(cardspec, TDA1004X_CONFC4, 1, 0)) !=0) return result; // wake up the chip
	   	inittab = tda10046_inittab;
		initcount = sizeof(tda10046_inittab) / 2;
		break;

	default:
	   	return -EIO;
	}

   	// upload inittable
	for(i=0; i< initcount; i+=2) {
		if ((result = cardspec->demod_write(cardspec, inittab[i], inittab+i+1, 1)) !=0) return result;
	}

	// final setup
	if (cardspec->demod_type == FE_TYPE_TDA10045H) tda10046h_set_bandwidth(cardspec, BANDWIDTH_8_MHZ);
	if ((result = tda1004x_write_mask(cardspec, TDA1004X_CONFC1, 0x40, cardspec->inversion ? 0x40 : 0)) !=0) return result; // toggle inversion setting

	return 0;
}

static int tda1004x_encode_fec(int fec)
{
	// convert known FEC values
	switch (fec) {
	case FEC_1_2:
		return 0;
	case FEC_2_3:
		return 1;
	case FEC_3_4:
		return 2;
	case FEC_5_6:
		return 3;
	case FEC_7_8:
		return 4;
	}

	// unsupported
	return -EINVAL;
}

static int tda1004x_decode_fec(int tdafec)
{
	// convert known FEC values
	switch (tdafec) {
	case 0:
		return FEC_1_2;
	case 1:
		return FEC_2_3;
	case 2:
		return FEC_3_4;
	case 3:
		return FEC_5_6;
	case 4:
		return FEC_7_8;
	}

	// unsupported
	return -1;
}

static int tda1004x_set_fe(struct tda1004x_cardspec* cardspec,
			   struct dvb_frontend_parameters *fe_params)
{
	int tmp;
	int inversion;

	dprintk("%s\n", __FUNCTION__);

	// set PLL
	tda1004x_enable_tuner_i2c(cardspec);
	cardspec->pll_set(cardspec, fe_params);
	tda1004x_disable_tuner_i2c(cardspec);   

	// Hardcoded to use auto as much as possible
	// The TDA10045 is very unreliable if AUTO mode is _not_ used. I have not
	// yet tested the TDA10046 to see if this issue has been fixed
	fe_params->u.ofdm.code_rate_HP = FEC_AUTO;
	fe_params->u.ofdm.guard_interval = GUARD_INTERVAL_AUTO;
	fe_params->u.ofdm.transmission_mode = TRANSMISSION_MODE_AUTO;

	// Set standard params.. or put them to auto
	if ((fe_params->u.ofdm.code_rate_HP == FEC_AUTO) ||
	    (fe_params->u.ofdm.code_rate_LP == FEC_AUTO) ||
	    (fe_params->u.ofdm.constellation == QAM_AUTO) ||
	    (fe_params->u.ofdm.hierarchy_information == HIERARCHY_AUTO)) {
		tda1004x_write_mask(cardspec, TDA1004X_AUTO, 1, 1);	// enable auto
		tda1004x_write_mask(cardspec, TDA1004X_IN_CONF1, 0x03, 0);	// turn off constellation bits
		tda1004x_write_mask(cardspec, TDA1004X_IN_CONF1, 0x60, 0);	// turn off hierarchy bits
		tda1004x_write_mask(cardspec, TDA1004X_IN_CONF2, 0x3f, 0);	// turn off FEC bits
	} else {
		tda1004x_write_mask(cardspec, TDA1004X_AUTO, 1, 0);	// disable auto

		// set HP FEC
		tmp = tda1004x_encode_fec(fe_params->u.ofdm.code_rate_HP);
		if (tmp < 0) return tmp;
		tda1004x_write_mask(cardspec, TDA1004X_IN_CONF2, 7, tmp);

		// set LP FEC
		tmp = tda1004x_encode_fec(fe_params->u.ofdm.code_rate_LP);
		if (tmp < 0) return tmp;
		tda1004x_write_mask(cardspec, TDA1004X_IN_CONF2, 0x38, tmp << 3);

		// set constellation
		switch (fe_params->u.ofdm.constellation) {
		case QPSK:
			tda1004x_write_mask(cardspec, TDA1004X_IN_CONF1, 3, 0);
			break;

		case QAM_16:
			tda1004x_write_mask(cardspec, TDA1004X_IN_CONF1, 3, 1);
			break;

		case QAM_64:
			tda1004x_write_mask(cardspec, TDA1004X_IN_CONF1, 3, 2);
			break;

		default:
			return -EINVAL;
		}

		// set hierarchy
		switch (fe_params->u.ofdm.hierarchy_information) {
		case HIERARCHY_NONE:
			tda1004x_write_mask(cardspec, TDA1004X_IN_CONF1, 0x60, 0 << 5);
			break;

		case HIERARCHY_1:
			tda1004x_write_mask(cardspec, TDA1004X_IN_CONF1, 0x60, 1 << 5);
			break;

		case HIERARCHY_2:
			tda1004x_write_mask(cardspec, TDA1004X_IN_CONF1, 0x60, 2 << 5);
			break;

		case HIERARCHY_4:
			tda1004x_write_mask(cardspec, TDA1004X_IN_CONF1, 0x60, 3 << 5);
			break;

		default:
			return -EINVAL;
		}
	}

	// set bandwidth
	switch(cardspec->demod_type) {
	case FE_TYPE_TDA10045H:
		tda10045h_set_bandwidth(cardspec, fe_params->u.ofdm.bandwidth);
		break;

	case FE_TYPE_TDA10046H:
		tda10046h_set_bandwidth(cardspec, fe_params->u.ofdm.bandwidth);
		break;
	}

	// need to invert the inversion if requested
	inversion = fe_params->inversion;
	if (cardspec->inversion) {
		inversion = inversion ? INVERSION_OFF : INVERSION_ON;
	}

	// set inversion
	switch (inversion) {
	case INVERSION_OFF:
		tda1004x_write_mask(cardspec, TDA1004X_CONFC1, 0x20, 0);
		break;

	case INVERSION_ON:
		tda1004x_write_mask(cardspec, TDA1004X_CONFC1, 0x20, 0x20);
		break;

	default:
		return -EINVAL;
	}

	// set guard interval
	switch (fe_params->u.ofdm.guard_interval) {
	case GUARD_INTERVAL_1_32:
		tda1004x_write_mask(cardspec, TDA1004X_AUTO, 2, 0);
		tda1004x_write_mask(cardspec, TDA1004X_IN_CONF1, 0x0c, 0 << 2);
		break;

	case GUARD_INTERVAL_1_16:
		tda1004x_write_mask(cardspec, TDA1004X_AUTO, 2, 0);
		tda1004x_write_mask(cardspec, TDA1004X_IN_CONF1, 0x0c, 1 << 2);
		break;

	case GUARD_INTERVAL_1_8:
		tda1004x_write_mask(cardspec, TDA1004X_AUTO, 2, 0);
		tda1004x_write_mask(cardspec, TDA1004X_IN_CONF1, 0x0c, 2 << 2);
		break;

	case GUARD_INTERVAL_1_4:
		tda1004x_write_mask(cardspec, TDA1004X_AUTO, 2, 0);
		tda1004x_write_mask(cardspec, TDA1004X_IN_CONF1, 0x0c, 3 << 2);
		break;

	case GUARD_INTERVAL_AUTO:
		tda1004x_write_mask(cardspec, TDA1004X_AUTO, 2, 2);
		tda1004x_write_mask(cardspec, TDA1004X_IN_CONF1, 0x0c, 0 << 2);
		break;

	default:
		return -EINVAL;
	}

	// set transmission mode
	switch (fe_params->u.ofdm.transmission_mode) {
	case TRANSMISSION_MODE_2K:
		tda1004x_write_mask(cardspec, TDA1004X_AUTO, 4, 0);
		tda1004x_write_mask(cardspec, TDA1004X_IN_CONF1, 0x10, 0 << 4);
		break;

	case TRANSMISSION_MODE_8K:
		tda1004x_write_mask(cardspec, TDA1004X_AUTO, 4, 0);
		tda1004x_write_mask(cardspec, TDA1004X_IN_CONF1, 0x10, 1 << 4);
		break;

	case TRANSMISSION_MODE_AUTO:
		tda1004x_write_mask(cardspec, TDA1004X_AUTO, 4, 4);
		tda1004x_write_mask(cardspec, TDA1004X_IN_CONF1, 0x10, 0);
		break;

	default:
		return -EINVAL;
	}

	// start the lock
	switch(cardspec->demod_type) {
	case FE_TYPE_TDA10045H:
		tda1004x_write_mask(cardspec, TDA1004X_CONFC4, 8, 8);
		tda1004x_write_mask(cardspec, TDA1004X_CONFC4, 8, 0);
		msleep(10);
		break;

	case FE_TYPE_TDA10046H:
		tda1004x_write_mask(cardspec, TDA1004X_AUTO, 0x40, 0x40);
		msleep(10);
		break;
	}

	return 0;
}

static int tda1004x_get_fe(struct tda1004x_cardspec* cardspec, struct dvb_frontend_parameters *fe_params)
{
	u8 val;

	dprintk("%s\n", __FUNCTION__);

	// inversion status
	fe_params->inversion = INVERSION_OFF;
	if (cardspec->demod_read(cardspec, TDA1004X_CONFC1, &val, 1)) return -EIO;
	if (val & 0x20) {
		fe_params->inversion = INVERSION_ON;
	}
	if (cardspec->inversion) {
		fe_params->inversion = fe_params->inversion ? INVERSION_OFF : INVERSION_ON;
	}

	// bandwidth
	switch(cardspec->demod_type) {
	case FE_TYPE_TDA10045H:
		if (cardspec->demod_read(cardspec, TDA10045H_WREF_LSB, &val, 1)) return -EIO;
		switch(val) {
		case 0x14:
			fe_params->u.ofdm.bandwidth = BANDWIDTH_8_MHZ;
			break;
		case 0xdb:
			fe_params->u.ofdm.bandwidth = BANDWIDTH_7_MHZ;
			break;
		case 0x4f:
			fe_params->u.ofdm.bandwidth = BANDWIDTH_6_MHZ;
			break;
		}
		break;

	case FE_TYPE_TDA10046H:
		if (cardspec->demod_read(cardspec, TDA10046H_TIME_WREF1, &val, 1)) return -EIO;
		switch(val) {
		case 0x60:
			fe_params->u.ofdm.bandwidth = BANDWIDTH_8_MHZ;
			break;
		case 0x6e:
			fe_params->u.ofdm.bandwidth = BANDWIDTH_7_MHZ;
			break;
		case 0x80:
			fe_params->u.ofdm.bandwidth = BANDWIDTH_6_MHZ;
			break;
		}
		break;
	}

	// FEC
	if (cardspec->demod_read(cardspec, TDA1004X_OUT_CONF2, &val, 1)) return -EIO;
	fe_params->u.ofdm.code_rate_HP = tda1004x_decode_fec(val & 7);
	fe_params->u.ofdm.code_rate_LP = tda1004x_decode_fec((val >> 3) & 7);

	// everything else uses OUT_CONF1
	if (cardspec->demod_read(cardspec, TDA1004X_OUT_CONF1, &val, 1)) return -EIO;
   
	// constellation
	switch (val & 3) {
	case 0:
		fe_params->u.ofdm.constellation = QPSK;
		break;
	case 1:
		fe_params->u.ofdm.constellation = QAM_16;
		break;
	case 2:
		fe_params->u.ofdm.constellation = QAM_64;
		break;
	}

	// transmission mode
	fe_params->u.ofdm.transmission_mode = TRANSMISSION_MODE_2K;
	if (val & 0x10) {
		fe_params->u.ofdm.transmission_mode = TRANSMISSION_MODE_8K;
	}

	// guard interval
	switch ((val & 0x0c) >> 2) {
	case 0:
		fe_params->u.ofdm.guard_interval = GUARD_INTERVAL_1_32;
		break;
	case 1:
		fe_params->u.ofdm.guard_interval = GUARD_INTERVAL_1_16;
		break;
	case 2:
		fe_params->u.ofdm.guard_interval = GUARD_INTERVAL_1_8;
		break;
	case 3:
		fe_params->u.ofdm.guard_interval = GUARD_INTERVAL_1_4;
		break;
	}

	// hierarchy
	switch ((val & 0x60) >> 5) {
	case 0:
		fe_params->u.ofdm.hierarchy_information = HIERARCHY_NONE;
		break;
	case 1:
		fe_params->u.ofdm.hierarchy_information = HIERARCHY_1;
		break;
	case 2:
		fe_params->u.ofdm.hierarchy_information = HIERARCHY_2;
		break;
	case 3:
		fe_params->u.ofdm.hierarchy_information = HIERARCHY_4;
		break;
	}

	return 0;
}

static int tda1004x_read_status(struct tda1004x_cardspec* cardspec, fe_status_t * fe_status)
{
	int cber;
	int vber;
	u8 val;

	dprintk("%s\n", __FUNCTION__);

	// read status
	if (cardspec->demod_read(cardspec, TDA1004X_STATUS_CD, &val, 1)) return -EIO;

	// decode
	*fe_status = 0;
	if (val & 4) *fe_status |= FE_HAS_SIGNAL;
	if (val & 2) *fe_status |= FE_HAS_CARRIER;
	if (val & 8) *fe_status |= FE_HAS_VITERBI | FE_HAS_SYNC | FE_HAS_LOCK;

	// if we don't already have VITERBI (i.e. not LOCKED), see if the viterbi
	// is getting anything valid
	if (!(*fe_status & FE_HAS_VITERBI)) {
		// read the CBER
		if (cardspec->demod_read(cardspec, TDA1004X_CBER_LSB, &val, 1)) return -EIO;
		cber = val;
		if (cardspec->demod_read(cardspec, TDA1004X_CBER_MSB, &val, 1)) return -EIO;
		cber |= (val << 8);
		cardspec->demod_read(cardspec, TDA1004X_CBER_RESET, &val, 1);

		if (cber != 65535) {
			*fe_status |= FE_HAS_VITERBI;
		}
	}

	// if we DO have some valid VITERBI output, but don't already have SYNC
	// bytes (i.e. not LOCKED), see if the RS decoder is getting anything valid.
	if ((*fe_status & FE_HAS_VITERBI) && (!(*fe_status & FE_HAS_SYNC))) {
		// read the VBER
		if (cardspec->demod_read(cardspec, TDA1004X_VBER_LSB, &val, 1)) return -EIO;
		vber = val;
		if (cardspec->demod_read(cardspec, TDA1004X_VBER_MID, &val, 1)) return -EIO;
		vber |= (val << 8);
		if (cardspec->demod_read(cardspec, TDA1004X_VBER_MSB, &val, 1)) return -EIO;
		vber |= ((val << 16) & 0x0f);
		cardspec->demod_read(cardspec, TDA1004X_CVBER_LUT, &val, 1);

		// if RS has passed some valid TS packets, then we must be
		// getting some SYNC bytes
		if (vber < 16632) {
			*fe_status |= FE_HAS_SYNC;
		}
	}

	// success
	dprintk("%s: fe_status=0x%x\n", __FUNCTION__, *fe_status);
	return 0;
}

static int tda1004x_read_signal_strength(struct tda1004x_cardspec* cardspec, u16 * signal)
{
	u8 val;
	int reg = 0;

	dprintk("%s\n", __FUNCTION__);

	// determine the register to use
	switch(cardspec->demod_type) {
	case FE_TYPE_TDA10045H:
		reg = TDA10045H_S_AGC;
		break;

	case FE_TYPE_TDA10046H:
		reg = TDA10046H_AGC_IF_LEVEL;
		break;
	}

	// read it
	if (cardspec->demod_read(cardspec, reg, &val, 1)) return -EIO;

	*signal = (val << 8) | val;
	dprintk("%s: signal=0x%x\n", __FUNCTION__, *signal);
	return 0;
}

static int tda1004x_read_snr(struct tda1004x_cardspec* cardspec, u16 * snr)
{
	u8 val;

	dprintk("%s\n", __FUNCTION__);

	// read it
	if (cardspec->demod_read(cardspec, TDA1004X_SNR, &val, 1)) return -EIO;
	if (val) {
		val = 255 - val;
	}

	*snr = ((val << 8) | val);
	dprintk("%s: snr=0x%x\n", __FUNCTION__, *snr);
	return 0;
}

static int tda1004x_read_ucblocks(struct tda1004x_cardspec* cardspec, u32* ucblocks)
{
	int tmp;
	int tmp2 = 0;
	int counter;
   	u8 val;

	dprintk("%s\n", __FUNCTION__);

	// read the UCBLOCKS and reset
	counter = 0;
	if (cardspec->demod_read(cardspec, TDA1004X_UNCOR, &val, 1)) return -EIO;
	tmp = val & 0x7f;
	while (counter++ < 5) {
		tda1004x_write_mask(cardspec, TDA1004X_UNCOR, 0x80, 0);
		tda1004x_write_mask(cardspec, TDA1004X_UNCOR, 0x80, 0);
		tda1004x_write_mask(cardspec, TDA1004X_UNCOR, 0x80, 0);

		if (cardspec->demod_read(cardspec, TDA1004X_UNCOR, &val, 1)) return -EIO;
		tmp2 &= (val & 0x7f);
		if ((tmp2 < tmp) || (tmp2 == 0))
			break;
	}

	if (tmp != 0x7f) {
		*ucblocks = tmp;
	} else {
		*ucblocks = 0xffffffff;
	}
	dprintk("%s: ucblocks=0x%x\n", __FUNCTION__, *ucblocks);
	return 0;
}

static int tda1004x_read_ber(struct tda1004x_cardspec* cardspec, u32* ber)
{
   	u8 val;

	dprintk("%s\n", __FUNCTION__);

	// read it in
	if ((cardspec->demod_read(cardspec, TDA1004X_CBER_LSB, &val, 1)) !=0) return -EIO;
	*ber = val << 1;
	if ((cardspec->demod_read(cardspec, TDA1004X_CBER_MSB, &val, 1)) !=0) return -EIO;
	*ber |= (val << 9);
	cardspec->demod_read(cardspec, TDA1004X_CBER_RESET, &val, 1);

	dprintk("%s: ber=0x%x\n", __FUNCTION__, *ber);
	return 0;
}

static int tda1004x_sleep(struct tda1004x_cardspec* cardspec)
{
	switch(cardspec->demod_type) {
	case FE_TYPE_TDA10045H:
		tda1004x_write_mask(cardspec, TDA1004X_CONFADC1, 0x10, 0x10);
		break;

	case FE_TYPE_TDA10046H:
		tda1004x_write_mask(cardspec, TDA1004X_CONFC4, 1, 1);
		break;
	}

	return 0;
}

static int _tda1004x_init(struct tda1004x_cardspec* cardspec)
{
	int result;
	struct dvb_frontend_info* fe_info;
   
	switch(cardspec->demod_type) {
	case FE_TYPE_TDA10045H:
		if ((result = tda10045_fwupload(cardspec)) !=0) return result;
		fe_info = &tda10045h_info;
		break;

	case FE_TYPE_TDA10046H:
		if ((result = tda10046_fwupload(cardspec)) !=0) return result;
		fe_info = &tda10046h_info;
		break;

	default:
		return -EIO;
	}

	tda1004x_enable_tuner_i2c(cardspec);
	cardspec->pll_init(cardspec);
	tda1004x_disable_tuner_i2c(cardspec);   

	return dvb_register_frontend(cardspec->ioctl, cardspec->adapter, cardspec->data, fe_info, THIS_MODULE);
}

int tda1004x_ioctl(struct tda1004x_cardspec* cardspec, struct dvb_frontend *fe, unsigned int cmd, void *arg)
{
	int status = 0;

	dprintk("%s: cmd=0x%x\n", __FUNCTION__, cmd);

	switch (cmd) {
	case FE_GET_INFO:
		switch(cardspec->demod_type) {
		case FE_TYPE_TDA10045H:
			memcpy(arg, &tda10045h_info, sizeof(struct dvb_frontend_info));
			break;

		case FE_TYPE_TDA10046H:
			memcpy(arg, &tda10046h_info, sizeof(struct dvb_frontend_info));
			break;
		}
		break;

	case FE_READ_STATUS:
		return tda1004x_read_status(cardspec, (fe_status_t *) arg);

	case FE_READ_BER:
		return tda1004x_read_ber(cardspec, (u32 *) arg);

	case FE_READ_SIGNAL_STRENGTH:
		return tda1004x_read_signal_strength(cardspec, (u16 *) arg);

	case FE_READ_SNR:
		return tda1004x_read_snr(cardspec, (u16 *) arg);

	case FE_READ_UNCORRECTED_BLOCKS:
		return tda1004x_read_ucblocks(cardspec, (u32 *) arg);

	case FE_SET_FRONTEND:
		return tda1004x_set_fe(cardspec, (struct dvb_frontend_parameters*) arg);

	case FE_GET_FRONTEND:
		return tda1004x_get_fe(cardspec, (struct dvb_frontend_parameters*) arg);

	case FE_SLEEP:
		cardspec->initialised = 0;
		return tda1004x_sleep(cardspec);

	case FE_INIT:

		// don't bother reinitialising
		if (cardspec->initialised) return 0;

		// OK, perform initialisation
		if ((status = tda1004x_init(cardspec)) !=0) return status;
		cardspec->initialised = 1;
		return 0;

	case FE_GET_TUNE_SETTINGS:
	{
		struct dvb_frontend_tune_settings* fesettings = (struct dvb_frontend_tune_settings*) arg;
		fesettings->min_delay_ms = 800;
		fesettings->step_size = 166667;
		fesettings->max_drift = 166667*2;
		return 0;
	}

	default:
		return -EOPNOTSUPP;
	}

	return 0;
}

int tda10045_init(struct tda1004x_cardspec* cardspec,
		  int (*ioctl) (struct dvb_frontend *frontend,
				unsigned int cmd, void *arg),
		  struct dvb_adapter *dvb_adapter,
		  void *data)
{
	int result;
	u8 chipid;

	dprintk("%s\n", __FUNCTION__);

	result = cardspec->demod_read(cardspec, TDA1004X_CHIPID, &chipid, 1);
	if (result || (chipid != 0x25)) return -ENODEV;

	cardspec->demod_type = FE_TYPE_TDA10045H;
	cardspec->initialised = 0;
   	cardspec->ioctl = ioctl;
	cardspec->adapter = dvb_adapter;
	cardspec->data = data;
	printk("tda1004x: Detected Philips TDA10045H.\n");
   
	return _tda1004x_init(cardspec);
}

int tda10046_init(struct tda1004x_cardspec* cardspec,
		  int (*ioctl) (struct dvb_frontend *frontend,
				unsigned int cmd, void *arg),
		  struct dvb_adapter *dvb_adapter,
		  void *data)
{
	int result;
	u8 chipid;

	dprintk("%s\n", __FUNCTION__);

	result = cardspec->demod_read(cardspec, TDA1004X_CHIPID, &chipid, 1);
	if (result || (chipid != 0x46)) return -ENODEV;

	cardspec->demod_type = FE_TYPE_TDA10046H;
    	cardspec->initialised = 0;
	cardspec->ioctl = ioctl;
	cardspec->adapter = dvb_adapter;
	cardspec->data = data;
	printk("tda1004x: Detected Philips TDA10046H.\n");

	return _tda1004x_init(cardspec);
}

int tda1004x_shutdown(struct tda1004x_cardspec* cardspec)
{
	tda1004x_sleep(cardspec);

	return dvb_unregister_frontend(cardspec->ioctl, cardspec->adapter);
}



MODULE_DESCRIPTION("Philips TDA10045H & TDA10046H DVB-T Frontend");
MODULE_AUTHOR("Andrew de Quincey & Robert Schlabbach");
MODULE_LICENSE("GPL");
